#coding:utf-8

# glottal voice source as input of Two Tubes Model of vocal tract
# Glottal Volume Velocity 
# based on A.E.Rosenberg's formula as Glottal Volume Velocity
#
# derived from https://github.com/shun60s/Vocal-Tube-Model

import numpy as np
from matplotlib import pyplot as plt

# Check version
#  Python 3.6.4 on win32 (Windows 10)
#  numpy 1.14.0 
#  matplotlib  2.1.1


class Class_Glottal(object):
	def __init__(self, tclosed=5.0, trise=6.0, tfall=2.0, sampling_rate=48000):
		# initalize
		self.tclosed=tclosed  # duration time of close state [mSec]
		self.trise=trise      # duration time of opening [mSec]
		self.tfall=tfall      # duration time of closing [mSec]
		self.sr= sampling_rate
		self.yg=self.make_one_plus()
		
	def make_one_plus(self,):
		# output yg
		self.N1=int( (self.tclosed / 1000.) * self.sr )
		self.N2=int( (self.trise / 1000.) * self.sr )
		self.N3=int( (self.tfall / 1000.) * self.sr )
		self.LL= self.N1+ self.N2 + self.N3
		yg=np.zeros(self.LL)
		#print ('Length= ', self.LL)
		for t0 in range(self.LL):
			if t0 < self.N1 :
				pass
			elif t0 <= (self.N2 + self.N1):
				yg[t0]= 0.5 * ( 1.0 - np.cos( ( np.pi / self.N2 ) * (t0 - self.N1)) )
			else:
				yg[t0]= np.cos( ( np.pi / ( 2.0 * self.N3 )) * ( t0 - (self.N2 + self.N1) )  )
		return yg

	def make_N_repeat(self, repeat_num=3):
		yg_repeat=np.zeros( len(self.yg) * repeat_num)
		for loop in range( repeat_num):
			yg_repeat[len(self.yg)*loop:len(self.yg)*(loop+1)]= self.yg
		return  yg_repeat
	
	def fone(self, f):
		# calculate one point of frequecny response
		xw= 2.0 * np.pi * f / self.sr
		yi=0.0
		yb=0.0
		for v in range (0,(self.N2 + self.N3)):
			yi+=  self.yg[self.N1 + v] * np.exp(-1j * xw * v)
			yb+=  self.yg[self.N1 + v]
		val= yi/yb
		return np.sqrt(val.real ** 2 + val.imag ** 2)
	
	def H0(self, freq_low=100, freq_high=5000, Band_num=256):
		# get Log scale frequecny response, from freq_low to freq_high, Band_num points
		amp=[]
		freq=[]
		bands= np.zeros(Band_num+1)
		fcl=freq_low * 1.0    # convert to float
		fch=freq_high * 1.0   # convert to float
		delta1=np.power(fch/fcl, 1.0 / (Band_num)) # Log Scale
		bands[0]=fcl
		#print ("i,band = 0", bands[0])
		for i in range(1, Band_num+1):
			bands[i]= bands[i-1] * delta1
			#print ("i,band =", i, bands[i]) 
		for f in bands:
			amp.append(self.fone(f) )
		return   np.log10(amp) * 20, bands # = amp value, freq list



if __name__ == '__main__':
	
	# instance
	glo=Class_Glottal()

	# draw
	fig = plt.figure()
	# draw one waveform
	plt.subplot(3,1,1)
	plt.xlabel('mSec')
	plt.ylabel('level')
	plt.title('Glottal Waveform')
	plt.plot( (np.arange(len(glo.yg)) * 1000.0 / glo.sr) , glo.yg)

	# draw frequecny response
	plt.subplot(3,1,2)
	plt.xlabel('Hz')
	plt.ylabel('dB')
	plt.title('Glottal frequecny response')
	amp, freq=glo.H0(freq_high=5000, Band_num=256)
	plt.plot(freq, amp)
	
	# draw repeated waveform
	yg_repeat=glo.make_N_repeat(repeat_num=3)
	plt.subplot(3,1,3)
	plt.xlabel('mSec')
	plt.ylabel('level')
	plt.title('Glottal repeated Waveform')
	plt.plot( (np.arange(len(yg_repeat)) * 1000.0 / glo.sr) , yg_repeat)
	
	#
	fig.tight_layout()
	plt.show()
	
#This file uses TAB



from string import Template
from random import randint
from math import floor

from IPython.display import display, HTML
try:
    from google.colab.output import eval_js
    HTML_fook = HTML
    display_hook = display


except:
    from IPython.display import IFrame
    from IPython.display import FileLink
    def eval_js(js):
        body = HTML(f"""<script>
        {js}
        </script>""")
        display(body)


    def HTML_fook(body):
        f = open('./tmp.html', 'w')
        f.write(body)
        f.close()


    def display_hook(html):
        display(IFrame('./tmp.html', width="100%", height="400px"))


def load_script():
    html_body = HTML_fook(LOAD_SCRIPT)
    display_hook(html_body)


from .face_html import HTML_FRAME, WORD, STYLES, SCRIPT

HTML_TEMPLATE = """
${style}
${frame}
${word}
${script}
<script>
${functions}
</script>
"""

class Face:
    """
    JupyterやColaboratoryの出力セルに，助手さんの顔を表示するためのクラス。
    """
    CENTER = 152
    EYEBLOW_POS = (52, 180)
    EYE_POS = (52, 220)
    MOUTH_POS = (142, 305)

    def __init__(self):
        """
        オブジェクトの初期化
        """
        self.message = ""
        self.eyeblow_position = [0, 0]
        self.eyeblow_degree = 0
        self.eye_position = [0, 0]
        self.eye_kind = 0
        self.mouth_position = [0, 0]
        self.mouth_kind = 0


    def show(self):
        """
        顔を表示する。
        """
        html_source = Template(HTML_TEMPLATE)
        args = {"style": STYLES, "frame": HTML_FRAME,
                "script": SCRIPT, "functions": "",
                "word": ""}
        fcs = ""
        if self.message:
            args["word"] = WORD
            fcs = f"say('{self.message}');"
        if self.eye_kind in [0, 1, 2]:
            fcs += f"setBackground('left_eye', 'eye_l_{self.eye_kind+1}.png');"
            fcs += f"setBackground('right_eye', 'eye_r_{self.eye_kind+1}.png');"
        if self.mouth_kind in [0, 1, 2]:
            fcs += f"setBackground('mouth', 'mouth_{self.mouth_kind+1}.png');"

        if self.eye_position != [0, 0]:
            fcs += f"setXPosition('left_eye', {self.CENTER-self.EYE_POS[0]+self.eye_position[0]});"
            fcs += f"setYPosition('left_eye', {self.EYE_POS[1]+self.eye_position[1]});"
            fcs += f"setXPosition('right_eye', {self.CENTER+self.EYE_POS[0]-self.eye_position[0]});"
            fcs += f"setYPosition('right_eye', {self.EYE_POS[1]+self.eye_position[1]});"

        if self.eyeblow_position != [0, 0]:
            fcs += f"setXPosition('left_eyeblow', {self.CENTER-self.EYEBLOW_POS[0]+self.eyeblow_position[0]});"
            fcs += f"setYPosition('left_eyeblow', {self.EYEBLOW_POS[1]+self.eyeblow_position[1]});"
            fcs += f"setXPosition('right_eyeblow', {self.CENTER+self.EYEBLOW_POS[0]-self.eyeblow_position[0]});"
            fcs += f"setYPosition('right_eyeblow', {self.EYEBLOW_POS[1]+self.eyeblow_position[1]});"

        if self.eyeblow_degree:
            fcs += f"rotate('left_eyeblow', {self.eyeblow_degree});"
            fcs += f"rotate('right_eyeblow', {-self.eyeblow_degree});"

        if self.mouth_position != [0, 0]:
            fcs += f"setXPosition('mouth', {self.MOUTH_POS[0]+self.mouth_position[0]});"
            fcs += f"setYPosition('mouth', {self.MOUTH_POS[1]+self.mouth_position[1]});"


        args["functions"] = fcs

        html_body = HTML_fook(html_source.substitute(**args))
        display_hook(html_body)


    def say(self, message):
        """
        顔にセリフを追加する。
        """
        self.message = message
        self.show()
        self.message = ""
    



#coding:utf-8

#
# Convert from Volume Velocity to Sound Pressure using High Pass Filter, for mouth radiation effect
#
# derived from https://github.com/shun60s/Vocal-Tube-Model

import numpy as np
from matplotlib import pyplot as plt

# Check version
#  Python 3.6.4 on win32 (Windows 10)
#  numpy 1.14.0 
#  matplotlib  2.1.1


class Class_HPF(object):
	def __init__(self, fc=1000, sampling_rate=48000):
		# initalize
		self.fc= fc # cut off frequency of High Pass Filter by unit is [Hz]
		self.sr= sampling_rate
		self.a, self.b = self.hpf1()
		
	def hpf1(self,):
		# primary digital filter
		a= np.ones(2)
		b= np.ones(2)
		wc= 2.0 * np.pi * self.fc / self.sr
		g0= np.tan( wc/2.0)
		a[1]= (g0 - 1.0) / ( g0 + 1.0)
		b[0]= 1.0 / (1.0 + g0)
		b[1]= -1.0 * b[0]
		return  a,b
		
	def iir1(self,x):
		# calculate iir filter: x is input, y is output
		# y[0]= b[0] * x[0]  + b[1] * x[-1]
		# y[0]= y[0] - a[1] * y[-1]
		y= np.zeros(len(x))
		for n in range(len(x)):
			for i in range(len(self.b)):
				if n - i >= 0:
					y[n] += self.b[i] * x[n - i]
			for j in range(1, len(self.a)):
				if n - j >= 0:
					y[n] -= self.a[j] * y[n - j]
		return y
		
	def fone(self, xw):
		# calculate one point of frequecny response
		f= xw / self.sr
		yi= self.b[0] + self.b[1] * np.exp(-2j * np.pi * f)
		yb= self.a[0] + self.a[1] * np.exp(-2j * np.pi * f)
		val= yi/yb
		return np.sqrt(val.real ** 2 + val.imag ** 2)

	def H0(self, freq_low=100, freq_high=5000, Band_num=256):
		# get Log scale frequecny response, from freq_low to freq_high, Band_num points
		amp=[]
		freq=[]
		bands= np.zeros(Band_num+1)
		fcl=freq_low * 1.0    # convert to float
		fch=freq_high * 1.0   # convert to float
		delta1=np.power(fch/fcl, 1.0 / (Band_num)) # Log Scale
		bands[0]=fcl
		#print ("i,band = 0", bands[0])
		for i in range(1, Band_num+1):
			bands[i]= bands[i-1] * delta1
			#print ("i,band =", i, bands[i]) 
		for f in bands:
			amp.append(self.fone(f))
		return   np.log10(amp) * 20, bands # = amp value, freq list



if __name__ == '__main__':
	
	# instance
	hpf=Class_HPF()
	
	# draw frequecny response
	plt.xlabel('Hz')
	plt.ylabel('dB')
	plt.title('High Pass Filter')
	amp, freq=hpf.H0(freq_high=5000, Band_num=256)
	plt.plot(freq, amp)
	plt.show()
	
#This file uses TAB



#coding:utf-8
# vowelgen
#
# 母音(アイウエ)を生成するプログラム。
#
# Copyright (c) 2021 Atsush Shibata(shivata@m-info.co.jp)
#
# Released under the MIT license.

import numpy as np
from scipy.io.wavfile import write as wavwrite
import IPython

from mpb_lib.vowelgen.twotube import Class_TwoTube
from mpb_lib.vowelgen.glottal import Class_Glottal
from mpb_lib.vowelgen.HPF import Class_HPF

glo=Class_Glottal()   # instance as glottal voice source
hpf=Class_HPF()       # instance for mouth radiation effect

def generate_vowel(L1_a, A1_a, L2_a, A2_a, length=50):
    twotube  =  Class_TwoTube(L1_a,L2_a,A1_a,A2_a)
    yg_repeat=glo.make_N_repeat(repeat_num=length) # input source of two tube model
    y2tm=twotube.process(yg_repeat)
    yout=hpf.iir1(y2tm)

    sampling_rate=48000
    wavwrite("vowel.wav", sampling_rate, (yout*2**15).astype(np.int16))
    return IPython.display.Audio("vowel.wav")


#coding:utf-8

#
# Three Tube Model, A python Class to calculate frequecny response and procee reflection transmission of resonance tube
#
# derived from https://github.com/shun60s/Vocal-Tube-Model


import numpy as np

# Check version
#  Python 3.6.4 on win32 (Windows 10)
#  numpy 1.14.0 



class Class_ThreeTube(object):
	def __init__(self, L3, A3, L1=9.0, L2=8.0, A1=1.0, A2=7.0, rg0=0.95, rl0=0.9 ,sampling_rate=48000):
		# initalize Tube length and Tube area
		self.L1= L1 # set list of 1st tube's length by unit is [cm]
		self.A1= A1 # set list of 1st tube's area by unit is [cm^2]
		self.L2= L2 # set list of 2nd tube's length by unit is [cm]
		self.A2= A2 # set list of 2nd tube's area by unit is [cm^2]
		self.L3= L3 # set list of 3rd tube's length by unit is [cm]
		self.A3= A3 # set list of 3rd tube's area by unit is [cm^2]
		C0=35000.0  # speed of sound in air, round 35000 cm/second
		self.sr= sampling_rate
		self.tu1=self.L1 / C0   # delay time in 1st tube
		self.tu2=self.L2 / C0   # delay time in 2nd tube
		self.tu3=self.L3 / C0   # delay time in 3rd tube
		self.r1=( self.A2 - self.A1) / ( self.A2 + self.A1)  # reflection coefficient between 1st tube and 2nd tube
		self.r2=( self.A3 - self.A2) / ( self.A3 + self.A2)  # reflection coefficient between 2nd tube and 3rd tube
		self.rg0=rg0 # rg is reflection coefficient between glottis and 1st tube
		self.rl0=rl0 # reflection coefficient between 3rd tube and mouth
		
	def fone(self, xw):
		# calculate one point of frequecny response
		yi= 0.5 * ( 1.0 + self.rg0 ) * ( 1.0 + self.r1)  * ( 1.0 + self.r2)  * ( 1.0 + self.rl0 ) * np.exp( -1.0j * ( self.tu1 + self.tu2 + self.tu3 ) * xw) 
		yb1= 1.0 + self.r1 * self.rg0 *  np.exp( -2.0j * self.tu1 * xw ) 
		yb1= yb1 + self.r2  * self.r1 *  np.exp( -2.0j * self.tu2 * xw ) 
		yb1= yb1 + self.rl0 * self.r2 *  np.exp( -2.0j * self.tu3 * xw ) 
		yb2=       self.r2  * self.rg0 *  np.exp( -2.0j * (self.tu1 + self.tu2) * xw ) 
		yb2= yb2 + self.rl0 * self.r1  *  np.exp( -2.0j * (self.tu2 + self.tu3) * xw ) 
		yb3= self.rl0 * self.r2 * self.r1 * self.rg0 *  np.exp( -2.0j * (self.tu1 + self.tu3) * xw )
		yb4= self.rl0 * self.rg0 * np.exp( -2.0j * (self.tu1 + self.tu2 + self.tu3) * xw ) 
		yb= yb1 + yb2 + yb3 + yb4
		val= yi/yb
		return np.sqrt(val.real ** 2 + val.imag ** 2)
		
	def H0(self, freq_low=100, freq_high=5000, Band_num=256):
		# get Log scale frequecny response, from freq_low to freq_high, Band_num points
		amp=[]
		freq=[]
		bands= np.zeros(Band_num+1)
		fcl=freq_low * 1.0    # convert to float
		fch=freq_high * 1.0   # convert to float
		delta1=np.power(fch/fcl, 1.0 / (Band_num)) # Log Scale
		bands[0]=fcl
		#print ("i,band = 0", bands[0])
		for i in range(1, Band_num+1):
			bands[i]= bands[i-1] * delta1
			#print ("i,band =", i, bands[i]) 
		for f in bands:
			amp.append(self.fone(f * 2.0 * np.pi))
		return   np.log10(amp) * 20, bands # = amp value, freq list

	def process(self, yg ):
		# process reflection transmission of resonance tube: yg is input, y2tm is output
		# three serial resonance tube
		#                      ---------------------
		#                      |                    |
		#   -------------------                      --------
		#   |                                               |
		#   |                                               |
		#   -------------------                     |--------
		#                      |                    |
		#                      ---------------------
		# reflection ratio
		#   rg                 r1                   r2                   rl0
		#   ya1---(forward)--->   yb1---(forward)---> yc1 ---(foward) --->
		#   <-----(backward)--ya2  <---(backward)---yb2  <---(backward)--yc2
		# input yg                                                      output y2tm
		# 
		#
		M1= round( self.tu1 * self.sr ) + 1  # for precision, higher sampling_rate is better
		M2= round( self.tu2 * self.sr ) + 1  # for precision, higher sampling_rate is better
		M3= round( self.tu3 * self.sr ) + 1  # for precision, higher sampling_rate is better
		M1= int(M1)
		M2= int(M2)
		M3= int(M3)
		ya1=np.zeros(M1)
		ya2=np.zeros(M1)
		yb1=np.zeros(M2)
		yb2=np.zeros(M2)
		yc1=np.zeros(M3)
		yc2=np.zeros(M3)
		y2tm=np.zeros(len(yg))
		
		for tc0 in range(len(yg)):
			for i in range((M1-1),0,-1): # process one step
				ya1[i]=ya1[i-1]
				ya2[i]=ya2[i-1]
			for i in range((M2-1),0,-1): # process one step
				yb1[i]=yb1[i-1]
				yb2[i]=yb2[i-1]	
			for i in range((M3-1),0,-1): # process one step
				yc1[i]=yc1[i-1]
				yc2[i]=yc2[i-1]	
			# calculate reflection
			ya1[0]= ((1. + self.rg0 ) / 2.) * yg[tc0] + self.rg0 * ya2[-1]
			ya2[0]= -1. * self.r1 *  ya1[-1]  +  ( 1. - self.r1 ) * yb2[-1]
			yb1[0]= ( 1 + self.r1 ) * ya1[-1] + self.r1 * yb2[-1]
			yb2[0]=  -1. * self.r2  * yb1[-1] + ( 1. - self.r2 ) * yc2[-1]
			yc1[0]= ( 1 + self.r2 ) * yb1[-1] + self.r2 * yc2[-1]
			yc2[0]=  -1. * self.rl0  * yc1[-1]
			y2tm[tc0]= (1 + self.rl0) * yc1[-1]

		return y2tm

if __name__ == '__main__':
	
	# Length & Area value, from problems 3.8 in "Digital Processing of Speech Signals" by L.R.Rabiner and R.W.Schafer
	#
	# /a/
	L1_a=9.0    # set list of 1st tube's length by unit is [cm]
	A1_a=1.0    # set list of 1st tube's area by unit is [cm^2]
	L2_a=8.0    # set list of 2nd tube's length by unit is [cm]
	A2_a=7.0    # set list of 2nd tube's area by unit is [cm^2]
	
	# /u/
	L1_u=10.0   # set list of 1st tube's length by unit is [cm]
	A1_u=7.0    # set list of 1st tube's area by unit is [cm^2]
	L2_u=7.0    # set list of 2nd tube's length by unit is [cm]
	A2_u=3.0    # set list of 2nd tube's area by unit is [cm^2]
	
	# /o/: L3,A3 is  extend factor to /a/ connecting as /u/
	L3_o= L2_a * (L2_u / L1_u)     # set list of 3rd tube's length by unit is [cm]
	A3_o= A2_a * (A2_u / A1_u)     # set list of 3rd tube's area by unit is [cm^2]
	
	# insatnce
	threetube_o  =  Class_ThreeTube(L3_o,A3_o)

	
#This file uses TAB



HTML_FRAME = """
    <div id="face">
        <div id="left_eyeblow"></div>
        <div id="right_eyeblow"></div>
        <div id="left_eye"></div>
        <div id="right_eye"></div>
        <div id="mouth"></div>
    </div>
"""

WORD = """
    <p class="balloon" id="word"></p>
"""


STYLES = """
<style>
    #face {
        background-image:url("https://github.com/shibats/mpb_samples/blob/main/assets/face/assistant_face.png?raw=true");
        width: 362px;
        height: 420px;
        margin-left: -100px;
        margin-top: -100px;
        transform: scale(0.5, 0.5);
    }
    #left_eyeblow {
        background-image:url("https://github.com/shibats/mpb_samples/blob/main/assets/face/eyeblow_l.png?raw=true");
        width: 54px;
        height: 20px;
        position: absolute;
        left: 100px;
        top: 180px;
    }
    #right_eyeblow {
        background-image:url("https://github.com/shibats/mpb_samples/blob/main/assets/face/eyeblow_r.png?raw=true");
        width: 54px;
        height: 20px;
        position: absolute;
        left: 205px;
        top: 180px;
    }
    #left_eye {
        background-image:url("https://github.com/shibats/mpb_samples/blob/main/assets/face/eye_l_1.png?raw=true");
        width: 56px;
        height: 38px;
        position: absolute;
        left: 100px;
        top: 220px;
    }
    #right_eye {
        background-image:url("https://github.com/shibats/mpb_samples/blob/main/assets/face/eye_r_1.png?raw=true");
        width: 56px;
        height: 38px;
        position: absolute;
        left: 205px;
        top: 220px;
    }
    #mouth {
        background-image:url("https://github.com/shibats/mpb_samples/blob/main/assets/face/mouth_1.png?raw=true");
        width: 78px;
        height: 60px;
        position: absolute;
        left: 140px;
        top: 305px;
    }

    .balloon {
        position: absolute;
        display: inline-block;
        margin: 1.5em 0 1.5em 15px;
        padding: 7px 10px;
        min-width: 120px;
        max-width: 95%;
        color: #555;
        font-size: 16px;
        background: #FFF;
        border: solid 3px #555;
        border-radius: 8px;
        box-sizing: border-box;
    }
    .balloon:before {
        content: "";
        position: absolute;
        top: 50%;
        left: -24px;
        margin-top: -12px;
        border: 12px solid transparent;
        border-right: 12px solid #FFF;
        z-index: 2;
    }
    .balloon:after {
        content: "";
        position: absolute;
        top: 50%;
        left: -30px;
        margin-top: -14px;
        border: 14px solid transparent;
        border-right: 14px solid #555;
        z-index: 1;
    }
    p.balloon  {
        margin-left: 200px;
        margin-top: -300px;
        padding: 0.5em;
        font-size: 24pt;
    }
</style>
"""

SCRIPT = """
<script>
    function setBackground(the_id, im) {
     var elementReference = document.getElementById(the_id);
     elementReference.style.backgroundImage = "url(https://github.com/shibats/mpb_samples/blob/main/assets/face/"+im+"?raw=true)";
     var backgroundImage = elementReference.style.backgroundImage;
    }

    function setXPosition(the_id, x) {
     var elementReference = document.getElementById(the_id);
     elementReference.style.left = x;
    }


    function setYPosition(the_id, x) {
     var elementReference = document.getElementById(the_id);
     elementReference.style.top = x;
    }


    function rotate(the_id, deg) {
     var elementReference = document.getElementById(the_id);
     elementReference.style.transform = "rotate("+deg+"deg)";
     var backgroundImage = elementReference.style.backgroundImage;
    }
    
    function say(msg) {
        var elementReference = document.getElementById("word");
        elementReference.innerHTML = msg;

    }

    /*
    setXPosition("left_eye", 105);
    setBackground("left_eye", "eye_l_1.png");
    setBackground("right_eye", "eye_r_1.png");
    setBackground("mouth", "mouth_1.png");
    rotate("left_eyeblow", -20);
    rotate("right_eyeblow", 20);
    say("オホホホホホ2");
    */
</script>
    
"""
